//System clock frequency definition
#define FCY (70000000)

//Device related headers
#include <p33EP128GP504.h>
#include <libpic30.h>
#include <dsp.h>


//Heater states definitions
#define INIT 0
#define COLD_CONDENSATION 1
#define COLD 2
#define QUITE_COLD 3
#define QUITE_WARM 4
#define PID_CLOSED 5
#define OVERHEAT 6

//Global variables used by module App_HeaterState for read:
extern unsigned int  App_Main_V_Nernst;
extern unsigned int  App_Main_V_RiH;
extern unsigned int  App_RiCalc_RiFilt;
extern unsigned long App_OS_SystemTime_Sec;
extern int           App_TempCalc_CellTemp;
extern unsigned int  App_DiagComDrv_IOCBI_Data_HeaterPID_TargetTemp;
extern unsigned int  App_DiagComDrv_IOCBI_Enable_HeaterPID_TargetTemp;
extern unsigned int  App_Main_V_Supply;
extern unsigned int  App_DiagComDrv_IOCBI_Data_HeaterPWMDutyCycle;
extern unsigned int  App_DiagComDrv_IOCBI_Enable_HeaterPWMDutyCycle;
extern unsigned char App_OS_TaskActive_100ms;
extern unsigned char App_OS_TaskActive_4ms;

//Global variables used by module App_HeaterState for write:
extern unsigned char App_HeaterControl_HeaterState;
extern unsigned char App_HeaterControl_OpPointReached;
extern unsigned int  App_HeaterControl_DesiredVoltage;
extern unsigned int  App_HeaterControl_DesiredPWMDutyCycle;

//External Calibrations definitions:
extern const unsigned int CAL_Ri_PID_CLOSED_Lth; //[0.1 Ohm/bit]
extern const unsigned int CAL_Ri_PID_CLOSED_Uth; //[0.1 Ohm/bit]
extern const unsigned int CAL_Ri_OVERHEAT_Lth; //[0.1 Ohm/bit]
extern const unsigned int CAL_Ri_OVERHEAT_Uth; //[0.1 Ohm/bit]
extern const unsigned int CAL_t_Condensation; //[1 sec/bit]
extern const unsigned int CAL_U_Condensation; //[1mV/bit]
extern const unsigned int CAL_U_Initial_Cold; //[1mV/bit]
extern const unsigned int CAL_dU_Cold; //[1mV/bit]
extern const unsigned int CAL_dU_QuiteCold; //[1mV/bit]
extern const unsigned int CAL_dU_QuiteWarm; //[1mV/bit]
extern const int          CAL_HeaterPID_TargetTemp; //[0.1C/bit]

//Data used by heater PID controller
extern tPID App_HeaterControl_PIDStc;
extern fractional App_HeaterControl_abcCoefficient[3];
extern fractional App_HeaterControl_controlHistory[3];
extern fractional App_HeaterControl_kCoeffs[3];


void App_HeaterControl (void)
{
    //Define the temporary RAM-cell for heater desired voltage signal
    unsigned int loc_DesiredVoltage;
    //Define a temporary RAM-cell for targer PWM duty cycle
    unsigned int loc_DesiredPWMDutyCycle;

    //*********************Compute the heater state****************************//
    switch (App_HeaterControl_HeaterState)
    {
        case INIT:
            App_HeaterControl_HeaterState = COLD;
            break;

        case COLD_CONDENSATION:
            //Set the condensation heater voltage
            App_HeaterControl_DesiredVoltage = CAL_U_Condensation;
            if (App_OS_SystemTime_Sec >= CAL_t_Condensation)
            {
                App_HeaterControl_HeaterState = COLD;
            }
            break;

        case COLD:
            if ((App_Main_V_Nernst >= 14500) && (App_OS_SystemTime_Sec < CAL_t_Condensation))
            {
                App_HeaterControl_HeaterState = COLD_CONDENSATION;
            }
            if (App_Main_V_Nernst < 14500)
            {
                App_HeaterControl_HeaterState = QUITE_COLD;
            }
            break;

        case QUITE_COLD:
            if (App_Main_V_Nernst >= 14500)
            {
                App_HeaterControl_HeaterState = COLD;
            }
            if (App_Main_V_RiH < 14500)
            {
                App_HeaterControl_HeaterState = QUITE_WARM;
            }
            break;

        case QUITE_WARM:
            if (App_Main_V_RiH >= 14500)
            {
                App_HeaterControl_HeaterState = QUITE_COLD;
            }
            if ((App_Main_V_RiH < 14500) && (App_RiCalc_RiFilt < CAL_Ri_PID_CLOSED_Lth))
            {
                App_HeaterControl_PIDStc.controlOutput = __builtin_divud(App_HeaterControl_DesiredVoltage, 10);
                App_HeaterControl_HeaterState = PID_CLOSED;
            }
            break;

        case PID_CLOSED:
            if ((App_Main_V_RiH < 14500) && (App_RiCalc_RiFilt > CAL_Ri_PID_CLOSED_Uth))
            {
                App_HeaterControl_HeaterState = QUITE_WARM;
            }
            if (App_Main_V_RiH >= 14500)
            {
                App_HeaterControl_HeaterState = QUITE_COLD;
            }
            if ((App_Main_V_RiH < 14500) && (App_RiCalc_RiFilt < CAL_Ri_OVERHEAT_Lth))
            {
                App_HeaterControl_HeaterState = OVERHEAT;
            }
            break;

        case OVERHEAT:
            //Cut off the heater voltage when overheat
            App_HeaterControl_DesiredVoltage = 0;
            if ((App_Main_V_RiH < 14500) && (App_RiCalc_RiFilt > CAL_Ri_OVERHEAT_Uth))
            {
                App_HeaterControl_HeaterState = PID_CLOSED;
            }
            if (App_Main_V_RiH >= 14500)
            {
                App_HeaterControl_HeaterState = QUITE_COLD;
            }
            break;
    }

    //***************************Heater control*********************************//
    //Warm-up control runs in 100ms cyclic
    if (App_OS_TaskActive_100ms)
    {
        //Define a local couter
        static unsigned int loc_cnt_Cold_increments;

        //Heater control in state "COLD"
        if (App_HeaterControl_HeaterState == COLD)
        {
            loc_DesiredVoltage = (CAL_U_Initial_Cold + loc_cnt_Cold_increments * CAL_dU_Cold);
            if (loc_DesiredVoltage > 12000)
            {
                loc_DesiredVoltage = 12000;
                //SET DTC_HEATER_COLD
            }
            else
            {
                loc_cnt_Cold_increments++;
            }
            App_HeaterControl_DesiredVoltage = loc_DesiredVoltage;
        }
        else
        {
            loc_cnt_Cold_increments = 0;
        }


        //Heater control in state "QUITE_COLD"
        if (App_HeaterControl_HeaterState == QUITE_COLD)
        {
            loc_DesiredVoltage = App_HeaterControl_DesiredVoltage + CAL_dU_QuiteCold;
            if (loc_DesiredVoltage > 12000)
            {
                loc_DesiredVoltage = 12000;
                //SET DTC_HEATER_QUITE_COLD
            }
            App_HeaterControl_DesiredVoltage = loc_DesiredVoltage;
        }

        //Heater control in state "QUITE_WARM"
        if (App_HeaterControl_HeaterState == QUITE_WARM)
        {
            loc_DesiredVoltage = App_HeaterControl_DesiredVoltage + CAL_dU_QuiteWarm;
            if (loc_DesiredVoltage > 12000)
            {
                loc_DesiredVoltage = 12000;
                //SET DTC_HEATER_QUITE_WARM
            }
            App_HeaterControl_DesiredVoltage = loc_DesiredVoltage;
        }
    }

    //PID CONTROLLER runs in 4ms cyclic, when control loop is closed
    if (App_OS_TaskActive_4ms && (App_HeaterControl_HeaterState == PID_CLOSED))
    {
        /*Load the target temperature for PID controller in normal mode*/
        App_HeaterControl_PIDStc.controlReference = CAL_HeaterPID_TargetTemp;
        /*Load the target temperature for PID controller in IOCBI mode*/
        if (App_DiagComDrv_IOCBI_Enable_HeaterPID_TargetTemp == 0x55AA)//IOCBI mode for PID-target temp enabled?
        {
            App_HeaterControl_PIDStc.controlReference = App_DiagComDrv_IOCBI_Data_HeaterPID_TargetTemp;
        }
        /*Load the actual measured sensor cell temperature*/
        App_HeaterControl_PIDStc.measuredOutput = App_TempCalc_CellTemp;
        /*Run the PID control function loaded from library libdsp-elf.a*/
        PID(&App_HeaterControl_PIDStc);
        
        //Cut off the negative half of bipolar PID result
        if (App_HeaterControl_PIDStc.controlOutput < (int)(0))
        {
            App_HeaterControl_PIDStc.controlOutput = (int)(0);
        }
        //Cut off the output greater than 12 Volts
        if (App_HeaterControl_PIDStc.controlOutput > (int)(1200))
        {
            App_HeaterControl_PIDStc.controlOutput = (int)(1200);
        }
        //Scale the PID result to 1mV/bit
        App_HeaterControl_DesiredVoltage = (App_HeaterControl_PIDStc.controlOutput * 10);
    }

    //Heater PWM is cut off when supply voltage is lower than 10V
    if (App_Main_V_Supply < 10000)
    {
        loc_DesiredPWMDutyCycle = 0;
        //Set DTC "Voltage too low" here
    }
    else
    {
    //Compute the target PWM duty cycle = Desired heater voltage / Supply voltage
    loc_DesiredPWMDutyCycle = __builtin_divud(
            (__builtin_muluu(App_HeaterControl_DesiredVoltage ,10000)), App_Main_V_Supply);
    }

    //IOCBI mode PWM control
    if (App_DiagComDrv_IOCBI_Enable_HeaterPWMDutyCycle == 0x55AA)//IOCBI mode for heater PWM enabled?
    {
        loc_DesiredPWMDutyCycle = App_DiagComDrv_IOCBI_Data_HeaterPWMDutyCycle;
    }

    //Set the desired duty cycle if less than 99.00%
    if (loc_DesiredPWMDutyCycle <= 9900)
    {
        App_HeaterControl_DesiredPWMDutyCycle = loc_DesiredPWMDutyCycle;
    }
    
    //Identify whether the heater operating point is reached
    if (
         //(App_HeaterControl_HeaterState == PID_CLOSED) &&
         ((App_TempCalc_CellTemp < 8500) && (App_TempCalc_CellTemp > 6300))
       )
    {
        App_HeaterControl_OpPointReached = 1;
    }
    else
    {
        App_HeaterControl_OpPointReached = 0;
    }
}