/*****************************************************************************/
// Hardware driver for:
// Output pin control for generation Ri measurement pulses
// ADC sampling and conversion of Nernst voltage, Supply voltage, Pump current
// DAC conversions for generation output AFR signal
// DAC conversions for generation of Ipump
// DMA connection to Diag com driver
// Output compare for Heater PWM generation
/*****************************************************************************/
//System clock frequency definition
#define FCY (70000000)

//Device related headers
#include <p33EP128GP504.h>
#include <libpic30.h>

//Hardware calibrations
#include "../CAL/HW_Calibrations.h"

//Hardware driver header
#include "../HWD/HWD.h"


/***************ADC1 HW Driver functions*******************/
int HWD_ADC1_convert_Nernst (void)
{
    //Define a local buffer for 3 ADC results
    unsigned int loc_ADCSum = 0;
    //Define a local RAM-cell for result
    int loc_ADCRes;
    //Local signals for indexing
    unsigned char i;
    //Set the ADC channel to scan a Nernst voltage
    AD1CHS0bits.CH0SA = 0; //Sample AN0 -> Nernst voltage
    
    //Make a three ADC samples and store them to the local buffer
    for (i = 0; i < 3; i++)
    {
        AD1CON1bits.SAMP = 1; //Start sampling and conversion
        while (IFS0bits.AD1IF == 0){}; //Wait until conversion is done
        loc_ADCSum += ADC1BUF0; //Add the result
        IFS0bits.AD1IF = 0; //Clear AD1 interrupt flag
    }
        
    //Compute the average
    loc_ADCRes = (loc_ADCSum / 3);

    //Compute and return the Nernst voltage [0.1 mV/bit] from average of ADC samples
    //V_AnIn[V] = ADCres * (Vref / 4095)
    //Nernst input ampifier ratio = 2.2, so: V_Nernst[V] = ADCres * (Vref / 2.2 / 4095)
    //Vref = 3.3 V, Output is in [0.1 mV/bit], so:
    //V_Nernst[0.1 mV/bit] = ADCBUF * (Vcc33 / 2.2 / 4095) - offset
    //-> V_Nernst = ADCBUF * 3.6630 - offset
    //-> V_Nernst = ADCBUF * 36630 / 10000 - offset
    int loc_Result = ((__builtin_divud((__builtin_muluu(loc_ADCRes, C_A_NernstAmp)), 1000)) + C_V_NernstAmpOffset);
    //Cut off the possible negative result
    if (loc_Result < 0)
    {
        loc_Result = 0;
    }
    //Make the result unsigned
    return((unsigned int)loc_Result);
}

int HWD_ADC1_convert_IPump (void)
{
    //Define a local buffer for 3 ADC results
    unsigned int loc_ADCSum = 0;
    //Define a local RAM-cell for result
    unsigned int loc_ADCRes;
    //Local signals for indexing
    unsigned char i;
    //Set the ADC channel to scan a Nernst voltage
    AD1CHS0bits.CH0SA = 1; //Sample AN1 -> Pump current
    
    //Make a three ADC samples and store them to the local buffer
    for (i = 0; i < 3; i++)
    {
        AD1CON1bits.SAMP = 1; //Start sampling and conversion
        while (IFS0bits.AD1IF == 0){}; //Wait until conversion is done
        loc_ADCSum += ADC1BUF0; //Add the result
        IFS0bits.AD1IF = 0; //Clear AD1 interrupt flag
    }
        
    //Store the median
    loc_ADCRes = (loc_ADCSum / 3);

    //Compute and return the Pump current [xxx uA/bit] from average of ADC samples
    //V_AnIn[V] = ADCres * (Vref / 4095) - c; where c is offset
    //Ipump = (V_AnIn / Aamp) / Rsens - c;
    //Ipump = ((ADCres * (Vref / 4095)) / Aamp) / Rsens - c
    //Ipump = ((ADCres * (Vcc33 / 4095)) / 3.928571) / 61.9 - c
    //Ipump[A] = ADCres * 3.313864e-06 - c
    //Ipump[uA] = ADCres * 33138 / 10000 - c
    int loc_Result = (__builtin_divud((__builtin_muluu(loc_ADCRes, C_A_PumpAmp)), 10000) - C_I_PumpOffsetIn);

    return(loc_Result);
}

int HWD_ADC1_convert_Supply (void)
{
    //Define a local buffer for 3 ADC results
    unsigned int loc_ADCSum = 0;
    //Define a local RAM-cell for result
    unsigned int loc_ADCRes;
    //Local signals for indexing
    unsigned char i;
    //Set the ADC channel to scan a Nernst voltage
    AD1CHS0bits.CH0SA = 2; //Sample AN2 -> Supply voltage
    
    //Make a three ADC samples and store them to the local buffer
    for (i = 0; i < 3; i++)
    {
        AD1CON1bits.SAMP = 1; //Start sampling and conversion
        while (IFS0bits.AD1IF == 0){}; //Wait until conversion is done
        loc_ADCSum += ADC1BUF0; //Add the result
        IFS0bits.AD1IF = 0; //Clear AD1 interrupt flag
    }
        
    //Store the median
    loc_ADCRes = (loc_ADCSum / 3);

    //Compute and return the supply voltage [1 mV/bit] from average of ADC samples:
    //V_AnIn[V] = ADCres * (Vref / 4095)
    //Voltage divider ratio = 10.9485, so: V_supply[V] = ADCBUF * (10.9485 * Vref / 4095)
    //Vref = 3.3 V, Output is in [mV], so: V_supply[mV] = ADCBUF * (10.9485 * 3300 / 4095)
    //-> V_supply[mV] = ADCBUF * 8.823
    //-> V_supply[mV] = ADCBUF * 8823 / 1000
    int loc_Result = __builtin_divud((__builtin_muluu(loc_ADCRes, C_A_SupplyAmp)), 1000);

    return(loc_Result);
}

/***************SPI2 HW Driver functions***************************************/
void HWD_OUTDAC_SetVoltage (unsigned int loc_Voltage) //1mV/bit
{
    //Compute the DAC input code from desired voltage:
    //DAC code = (4095 * Voltage)/5000
    unsigned int loc_DAC_Code;
    loc_DAC_Code = (__builtin_divud((__builtin_muluu(loc_Voltage, 4095)), C_V_Vcc5));
    //Add the four MSBits (DAC command-bits):
    //Bit 15: 0 -> Always 0
    //Bit 14: 0 -> No buffer
    //Bit 13: 1 -> Vout = 1xVref*D/4096
    //Bit 12: 1 -> Active mode
    //Send the data by SPI2
    SPI2_OUTDAC_CS = CS_ACTIVE; //Activate chip select pin for Ri DAC
    SPI2BUF = ((loc_DAC_Code & 0x0FFF)|0x3000);
    while (!SPI2STATbits.SPIRBF); //Wait to complete the transfer
    SPI2BUF; //Read the SPI buffer to clear SPIRBF flag
    SPI2_OUTDAC_CS = CS_INACTIVE; //Deactivate chip select pin for Ri DAC
}

void HWD_OUTDAC_ResetVoltage (void)
{
    SPI2_OUTDAC_CS = CS_ACTIVE; //Activate chip select pin for IPump DAC
    //This command sets the output voltage to 2.5 V and deactivates the DAC out-
    //put.
    SPI2BUF = 0x219E; //Write the data to be transmitted to the SPI buffer
    while (!SPI2STATbits.SPIRBF); //Wait to complete the transfer
    SPI2BUF; //Read the SPI buffer to clear SPIRBF flag
    SPI2_OUTDAC_CS = CS_INACTIVE; //Deactivate chip select pin for IPump DAC
}


void HWD_PumpDAC_SetUPump (unsigned int loc_DAC_Code)
{   //Add the four MSBits (DAC command-bits):
    //Bit 15: 0 -> Always 0
    //Bit 14: 0 -> No buffer
    //Bit 13: 1 -> Vout = 1xVref*D/4096
    //Bit 12: 1 -> Active mode
    //Send the data by SPI2
    SPI2_IPumpDAC_CS = CS_ACTIVE; //Activate chip select pin for IPump DAC
    SPI2BUF = ((loc_DAC_Code & 0x0FFF)|0x3000); //Write the data to be transmitted to the SPI buffer
    while (!SPI2STATbits.SPIRBF); //Wait to complete the transfer
    SPI2BUF; //Read the SPI buffer to clear SPIRBF flag
    SPI2_IPumpDAC_CS = CS_INACTIVE; //Deactivate chip select pin for IPump DAC
}

void HWD_PumpDAC_ResetUPump (void)
{
    SPI2_IPumpDAC_CS = CS_ACTIVE; //Activate chip select pin for IPump DAC
    //This command sets the output voltage to 2.5 V and deactivates the DAC out-
    //put.
    SPI2BUF = 0x27FF; //Write the data to be transmitted to the SPI buffer
    while (!SPI2STATbits.SPIRBF); //Wait to complete the transfer
    SPI2BUF; //Read the SPI buffer to clear SPIRBF flag
    SPI2_IPumpDAC_CS = CS_INACTIVE; //Deactivate chip select pin for IPump DAC
}

void HWD_HeaterPWM_SetPulse (unsigned int loc_DutyCycle)//[0.01% / bit]
{
    unsigned int loc_PulseOffset; //Time offset of rising edge of PWM-pulse, to
    //avoid edge-interferences when ADC samples

    //Compute the desired pulse length from duty cycle:
    //PWM Period = 28000 ticks
    //PulseLength [ticks] = (PWM Period / 100) * Desired_duty_cycle[%]
    //-> PulseLength = (28000 * Desired_duty_cycle[0.01% / bit]) / 10000
    unsigned int loc_PulseLength = __builtin_divud((__builtin_muluu(14, loc_DutyCycle)), 5);
    
    //Compute the time offset of heater pulse rising edge for avoiding
    //the collision of the PWM edges with ADC samplings:
    
    if ((loc_PulseLength > 3500) && (loc_PulseLength < 6300)) //50us && 90 us
    {
        loc_PulseOffset = 2800;//40us
    }
    else if (((loc_PulseLength > 14350) && (loc_PulseLength < 16100))//205us && 230 us
            ||
            ((loc_PulseLength > 23100) && (loc_PulseLength < 24850)))//330us && 355us
    {
        loc_PulseOffset = 1750;//25us
    }
    else
    {
        loc_PulseOffset = 0;
    }

    OC1CON1bits.OCM = 0; //Reset the OC1 module
    OC1TMR = 0;
    OC1R = (loc_PulseOffset + 1); //Addition of 1 is needed by HW - OC module can
    //not generate a pulse when OC1R = 0
    OC1RS = (loc_PulseLength + loc_PulseOffset + 1);
    //If the pulse length is equal to 0, OC module is not turned on to avoid
    //generation of a pulse.
    //If not zero, Double Compare Single-Shot mode starts (Low->High->Low mode)
    if (loc_PulseLength != 0){OC1CON1bits.OCM = 0b100;}
}

unsigned char HWD_CheckUARTRXBuffer (void)
{
    unsigned char ret = 0;
    //Check if DMA transfer of 8 bytes is complete
    if(IFS0bits.DMA1IF == 1)
    {
       //Clear the DMA transfer complete flag
       IFS0bits.DMA1IF = 0;
       ret = 1;
    }
    return(ret);
}